import { ReallySmallEvents } from "really-small-events";
import { InanoSQLConfig, InanoSQLFunction, InanoSQLQuery, InanoSQLDatabaseEvent, InanoSQLInstance, InanoSQLQueryBuilder, TableQueryResult, InanoSQLV1ConfigFn, InanoSQLDBConfig } from "./interfaces";
export { InanoSQLInstance };
export declare class nanoSQL implements InanoSQLInstance {
    dbs: {
        [id: string]: InanoSQLDBConfig;
    };
    selectedDB: string;
    indexTypes: {
        [type: string]: (value: any) => any;
    };
    version: number;
    functions: {
        [fnName: string]: InanoSQLFunction;
    };
    events: {
        [id: string]: {
            Core: {
                [path: string]: ReallySmallEvents;
            };
            [eventName: string]: {
                [path: string]: ReallySmallEvents;
            };
        };
    };
    planetRadius: number;
    selectedTable: any;
    txs: {
        [id: string]: {
            table: string;
            type: "put" | "del" | "idx-put" | "idx-del";
            data: any;
        }[];
    };
    constructor();
    getDB(id?: string): InanoSQLDBConfig;
    _rebuildFKs(): void;
    doFilter<T>(databaseID: string | undefined, filterName: string, args: T, complete: (result: T) => void, cancelled: (abortInfo: any) => void): void;
    getCache(id: string, args?: {
        offset: number;
        limit: number;
    }): any[];
    clearCache(id: string): boolean;
    clearTTL(primaryKey: any): Promise<any>;
    expires(primaryKey: any): Promise<any>;
    _checkTTL(): void;
    selectTable(table?: string | any[] | ((where?: any[] | ((row: {
        [key: string]: any;
    }, i?: number) => boolean)) => Promise<TableQueryResult>)): InanoSQLInstance;
    getPeers(): any;
    _initPlugins(config: InanoSQLConfig): Promise<any>;
    _saveTableIds(databaseID: string): Promise<any>;
    presetQuery(fn: string, args?: any): InanoSQLQueryBuilder;
    useDatabase(id: string): InanoSQLInstance;
    createDatabase(config?: InanoSQLConfig): Promise<any>;
    listDatabases(): string[];
    dropDatabase(id: string): Promise<any>;
    maybeCreateEventObject(id: string): void;
    connect(config?: InanoSQLConfig): Promise<any>;
    _initPeers(): void;
    every(args: {
        length: number;
        every?: number;
        offset?: number;
    }): number[];
    on(action: string, callBack: (event: InanoSQLDatabaseEvent) => void, selectTable?: string): void;
    off(action: string, callBack: (event: InanoSQLDatabaseEvent) => void, selectTable?: string): void;
    _refreshEventChecker(): InanoSQLInstance;
    getView(viewName: string, viewArgs: any): Promise<any>;
    doAction(actionName: string, actionArgs: any): Promise<any>;
    _doAV(AVType: "a" | "v", table: string, AVName: string, AVArgs: any): Promise<any>;
    query(action: string | ((nSQL: InanoSQLInstance) => InanoSQLQuery), args?: any): InanoSQLQueryBuilder;
    triggerQuery(databaseID: string | undefined, query: InanoSQLQuery, onRow: (row: any) => void, complete: () => void, error: (err: string) => void): void;
    triggerEvent(databaseID: string | undefined, eventData: InanoSQLDatabaseEvent, ignoreStarTable?: boolean): InanoSQLInstance;
    private _countTimers;
    saveCount(databaseID: string, tableName: string, complete?: (err?: any) => void): void;
    default(databaseID: string | undefined, replaceObj?: any, table?: string): {
        [key: string]: any;
    } | Error;
    rawDump(tables: string[], indexes: boolean, onRow: (table: string, row: {
        [key: string]: any;
    }) => void): Promise<any>;
    rawImport(tables: {
        [table: string]: {
            [key: string]: any;
        }[];
    }, indexes: boolean, onProgress?: (percent: number) => void): Promise<any>;
    disconnect(dbID?: string): Promise<unknown>;
    extend(scope: string, ...args: any[]): any | nanoSQL;
    loadJS(rows: {
        [key: string]: any;
    }[], onProgress?: (percent: number) => void, parallel?: boolean): Promise<any[]>;
    JSONtoCSV(json: any[], printHeaders?: boolean, useHeaders?: string[]): string;
    csvToArray(text: string): any[];
    CSVtoJSON(csv: string, rowMap?: (row: any) => any): any;
    loadCSV(csvString: string, rowMap?: (row: any) => any, onProgress?: (percent: number) => void, parallel?: boolean): Promise<any[]>;
}
export declare const nSQLv1Config: (doConfig: (nSQLv1: (table?: string | undefined) => InanoSQLV1ConfigFn) => void) => InanoSQLConfig;
export declare const nSQL: (table?: string | any[] | ((where?: any[] | ((row: {
    [key: string]: any;
}, i?: number | undefined) => boolean) | undefined) => Promise<TableQueryResult>) | undefined) => InanoSQLInstance;
